@@ -1,5 +1,10 @@
-{-# LANGUAGE OverloadedStrings, ScopedTypeVariables #-}
-{-# OPTIONS_GHC -fno-warn-orphans #-}
+{-# LANGUAGE CPP, DeriveGeneric, OverloadedStrings, ScopedTypeVariables #-}
+
+#if __GLASGOW_HASKELL__ >= 801
+{-# OPTIONS_GHC -Wno-orphans -Wno-unused-top-binds #-}
+#else
+{-# OPTIONS_GHC -fno-warn-orphans -fno-warn-unused-binds #-}
+#endif
 
 module Main
     ( main
@@ -10,20 +15,28 @@ import qualified Data.ByteString.Lazy as BL
 import qualified Data.ByteString.Lazy.Char8 as BL8
 import qualified Data.HashMap.Strict as HM
 import Data.Int
+import Data.Scientific (Scientific)
 import qualified Data.Text as T
 import qualified Data.Text.Lazy as LT
 import qualified Data.Vector as V
 import qualified Data.Foldable as F
 import Data.Word
+import Numeric.Natural
+import GHC.Generics (Generic)
 import Test.HUnit
 import Test.Framework as TF
 import Test.Framework.Providers.HUnit as TF
 import Test.QuickCheck
+import Test.QuickCheck.Instances ()
 import Test.Framework.Providers.QuickCheck2 as TF
 
 import Data.Csv hiding (record)
 import qualified Data.Csv.Streaming as S
 
+#if !MIN_VERSION_base(4,8,0)
+import Control.Applicative ((<$>), (<*>))
+#endif
+
 ------------------------------------------------------------------------
 -- Parse tests
 
@@ -111,7 +124,7 @@ positionalTests =
       , ("quoted",       [["\"abc\""]],      "\"\"\"abc\"\"\"\r\n")
       , ("quote",        [["a\"b"]],         "\"a\"\"b\"\r\n")
       , ("quotedQuote",  [["\"a\"b\""]],     "\"\"\"a\"\"b\"\"\"\r\n")
-      , ("leadingSpace", [[" abc"]],         "\" abc\"\r\n")
+      , ("leadingSpace", [[" abc"]],         " abc\r\n")
       , ("comma",        [["abc,def"]],      "\"abc,def\"\r\n")
       , ("twoFields",    [["abc","def"]],    "abc,def\r\n")
       , ("twoRecords",   [["abc"], ["def"]], "abc\r\ndef\r\n")
@@ -240,18 +253,6 @@ nameBasedTests =
 ------------------------------------------------------------------------
 -- Conversion tests
 
-instance Arbitrary B.ByteString where
-    arbitrary = B.pack `fmap` arbitrary
-
-instance Arbitrary BL.ByteString where
-    arbitrary = BL.fromChunks `fmap` arbitrary
-
-instance Arbitrary T.Text where
-    arbitrary = T.pack `fmap` arbitrary
-
-instance Arbitrary LT.Text where
-    arbitrary = LT.fromChunks `fmap` arbitrary
-
 -- A single column with an empty string is indistinguishable from an
 -- empty line (which we will ignore.) We therefore encode at least two
 -- columns.
@@ -291,11 +292,13 @@ conversionTests =
       , testProperty "Int16" (roundTrip :: Int16 -> Bool)
       , testProperty "Int32" (roundTrip :: Int32 -> Bool)
       , testProperty "Int64" (roundTrip :: Int64 -> Bool)
+      , testProperty "Natural" (roundTrip :: Natural -> Bool)
       , testProperty "Word" (roundTrip :: Word -> Bool)
       , testProperty "Word8" (roundTrip :: Word8 -> Bool)
       , testProperty "Word16" (roundTrip :: Word16 -> Bool)
       , testProperty "Word32" (roundTrip :: Word32 -> Bool)
       , testProperty "Word64" (roundTrip :: Word64 -> Bool)
+      , testProperty "Scientific" (roundTrip :: Scientific -> Bool)
       , testProperty "lazy ByteString"
         (roundTrip :: BL.ByteString -> Bool)
       , testProperty "Text" (roundTrip :: T.Text -> Bool)
@@ -321,25 +324,35 @@ conversionTests =
                             "Cam yiyebilirim, bana zararı dokunmaz.")
       ]
     , testGroup "Partial Decodes"
-      [ testCase "Int"     (partialDecode
-                            (parseField "12.7" :: Parser Int))
-      , testCase "Word"    (partialDecode
-                            (parseField "12.7" :: Parser Word))
-      , testCase "Double"  (partialDecode
-                            (parseField "1.0+" :: Parser Double))
-      , testCase "Integer" (partialDecode
-                            (parseField "1e6"  :: Parser Integer))
+      [ testCase "Int"         (partialDecode
+                                (parseField "12.7" :: Parser Int))
+      , testCase "Natural"     (partialDecode
+                                (parseField "12.7" :: Parser Natural))
+      , testCase "Word"        (partialDecode
+                                (parseField "12.7" :: Parser Word))
+      , testCase "Scientific"  (partialDecode
+                                (parseField "1.0+" :: Parser Scientific))
+      , testCase "Double"      (partialDecode
+                                (parseField "1.0+" :: Parser Double))
+      , testCase "Integer"     (partialDecode
+                                (parseField "1e6"  :: Parser Integer))
       ]
     , testGroup "Space trimming"
-      [ testCase "_Int"     (expect (parseField " 12"     :: Parser Int)    12)
-      , testCase "Int_"     (expect (parseField "12 "     :: Parser Int)    12)
-      , testCase "_Int_"    (expect (parseField " 12 "    :: Parser Int)    12)
-      , testCase "_Word"    (expect (parseField " 12"     :: Parser Word)   12)
-      , testCase "Word_"    (expect (parseField "12 "     :: Parser Word)   12)
-      , testCase "_Word_"   (expect (parseField " 12 "    :: Parser Word)   12)
-      , testCase "_Double"  (expect (parseField " 1.2e1"  :: Parser Double) 12)
-      , testCase "Double_"  (expect (parseField "1.2e1 "  :: Parser Double) 12)
-      , testCase "_Double_" (expect (parseField " 1.2e1 " :: Parser Double) 12)
+      [ testCase "_Int"         (expect (parseField " 12"     :: Parser Int)        12)
+      , testCase "Int_"         (expect (parseField "12 "     :: Parser Int)        12)
+      , testCase "_Int_"        (expect (parseField " 12 "    :: Parser Int)        12)
+      , testCase "_Natural"     (expect (parseField " 12"     :: Parser Natural)    12)
+      , testCase "Natural_"     (expect (parseField "12 "     :: Parser Natural)    12)
+      , testCase "_Natural_"    (expect (parseField " 12 "    :: Parser Natural)    12)
+      , testCase "_Word"        (expect (parseField " 12"     :: Parser Word)       12)
+      , testCase "Word_"        (expect (parseField "12 "     :: Parser Word)       12)
+      , testCase "_Word_"       (expect (parseField " 12 "    :: Parser Word)       12)
+      , testCase "_Scientific"  (expect (parseField " 1.2e1"  :: Parser Scientific) 12)
+      , testCase "Scientific_"  (expect (parseField "1.2e1 "  :: Parser Scientific) 12)
+      , testCase "_Scientific_" (expect (parseField " 1.2e1 " :: Parser Scientific) 12)
+      , testCase "_Double"      (expect (parseField " 1.2e1"  :: Parser Double)     12)
+      , testCase "Double_"      (expect (parseField "1.2e1 "  :: Parser Double)     12)
+      , testCase "_Double_"     (expect (parseField " 1.2e1 " :: Parser Double)     12)
       ]
     ]
 
@@ -370,6 +383,7 @@ instanceTests =
   [
     testGroup "Records instances"
     [ testCase "foldr Foldable" (expected @=? F.foldr (:) [] input)
+    , testCase "foldl' Foldable" (expected @=? F.foldl' (flip (:)) [] input)
     ]
   ]
   where
@@ -377,6 +391,53 @@ instanceTests =
       S.Cons (Right ("a" :: String)) (S.Nil Nothing BL8.empty))
     expected = ["a" :: String]
 
+------------------------------------------------------------------------
+-- Custom conversion option tests
+
+genericConversionTests :: [TF.Test]
+genericConversionTests =
+    [ testCase "headerOrder" (header ["column1", "column2", "column_3"] @=? hdrs)
+    , testCase "encode" (encodeDefaultOrderedByName sampleValues @?= sampleEncoding)
+    , testCase "decode" (Right (hdrs, V.fromList sampleValues) @=? decodeByName sampleEncoding)
+    , testProperty "roundTrip" rtProp
+    ]
+  where
+    hdrs = headerOrder (undefined :: SampleType)
+
+    sampleValues = [ SampleType ""      1     Nothing
+                   , SampleType "field" 99999 (Just 1.234)
+                   ]
+
+    sampleEncoding = "column1,column2,column_3\r\n,1,\r\nfield,99999,1.234\r\n"
+
+    rtProp :: [SampleType] -> Bool
+    rtProp vs = Right (hdrs, V.fromList vs)
+                == decodeByName (encodeDefaultOrderedByName vs)
+
+data SampleType = SampleType
+  { _column1  :: !T.Text
+  , column2   :: !Int
+  , _column_3 :: !(Maybe Double)
+  } deriving (Eq, Show, Read, Generic)
+
+sampleOptions :: Options
+sampleOptions = defaultOptions { fieldLabelModifier = rmUnderscore }
+  where
+    rmUnderscore ('_':str) = str
+    rmUnderscore str       = str
+
+instance ToNamedRecord SampleType where
+  toNamedRecord = genericToNamedRecord sampleOptions
+
+instance FromNamedRecord SampleType where
+  parseNamedRecord = genericParseNamedRecord sampleOptions
+
+instance DefaultOrdered SampleType where
+  headerOrder = genericHeaderOrder sampleOptions
+
+instance Arbitrary SampleType where
+  arbitrary = SampleType <$> arbitrary <*> arbitrary <*> arbitrary
+
 ------------------------------------------------------------------------
 -- Test harness
 
@@ -386,6 +447,7 @@ allTests = [ testGroup "positional" positionalTests
            , testGroup "conversion" conversionTests
            , testGroup "custom-options" customOptionsTests
            , testGroup "instances" instanceTests
+           , testGroup "generic-conversions" genericConversionTests
            ]
 
 main :: IO ()
